use crate::{config, key, schema, token, UIDCError};
use microrm::prelude::*;
use ring::signature::KeyPair;
use serde::{Deserialize, Serialize};

use super::session::SessionHelper;

mod api;

type Request = tide::Request<super::ServerStateWrapper>;

const AUTHORIZE_PATH: &'static str = "oidc/authorize";
const TOKEN_PATH: &'static str = "oidc/token";
const JWKS_PATH: &'static str = "oidc/jwks";
const DISCOVERY_PATH: &'static str = ".well-known/openid-configuration";

#[derive(serde::Serialize)]
pub enum OIDCErrorType {
    InvalidRequest,
    UnauthorizedClient,
    AccessDenied,
    UnsupportedResponseType,
    // InvalidScope,
    ServerError,
    // TemporarilyUnavailable,
}

/// error type,
pub struct OIDCError<'a>(OIDCErrorType, String, Option<&'a str>);

impl<'a> OIDCError<'a> {
    fn to_response(self) -> tide::Response {
        #[derive(Serialize)]
        struct ErrorOut<'a> {
            error: OIDCErrorType,
            error_description: String,
            state: Option<&'a str>,
        }

        let eo = ErrorOut {
            error: self.0,
            error_description: self.1,
            state: self.2,
        };

        tide::Response::builder(400)
            .body(serde_json::to_vec(&eo).unwrap())
            .build()
    }
}

impl<'a> From<microrm::Error> for OIDCError<'a> {
    fn from(value: microrm::Error) -> Self {
        Self(
            OIDCErrorType::ServerError,
            format!("Internal database error: {value}"),
            None,
        )
    }
}

#[derive(Deserialize)]
struct AuthorizeQueryParams {
    response_type: String,
    client_id: String,
    redirect_uri: String,
    scope: Option<String>,
}

fn do_code_authorize<'l, 's>(
    config: &config::Config,
    realm: &microrm::Stored<schema::Realm>,
    client: &microrm::Stored<schema::Client>,
    user: &microrm::Stored<schema::User>,
    scopes: impl Iterator<Item = &'l str>,
    redirect_uri: String,
    state: Option<&'s str>,
) -> Result<tide::Response, OIDCError<'s>> {
    let expiry =
        std::time::SystemTime::now() + std::time::Duration::from_secs(config.auth_code_expiry);

    let rng = ring::rand::SystemRandom::new();
    let raw_auth_code: [u8; 32] = ring::rand::generate(&rng)
        .map_err(|_| {
            OIDCError(
                OIDCErrorType::ServerError,
                format!("Failed to generate auth code."),
                state,
            )
        })?
        .expose();
    let encoded_auth_code = base64::encode_config(raw_auth_code, base64::URL_SAFE_NO_PAD);

    realm.auth_codes.insert(schema::AuthCode {
        realm: realm.id(),
        client: client.id(),
        user: user.id(),
        scopes: scopes
            .map(|x| x.to_owned())
            .collect::<Vec<_>>()
            .into_serialized(),
        expiry: expiry.into(),
        redirect_uri: redirect_uri.clone(),
        code: encoded_auth_code.clone(),
    })?;

    let new_params = [
        ("response_type", "code"),
        ("code", encoded_auth_code.as_str()),
    ]
    .into_iter()
    .chain(state.map(|v| ("state", v)));

    Ok(tide::Redirect::temporary(
        tide::http::Url::parse_with_params(redirect_uri.as_str(), new_params).map_err(|e| {
            OIDCError(
                OIDCErrorType::InvalidRequest,
                format!("could not parse redirect_uri as a URL: {e}"),
                None,
            )
        })?,
    )
    .into())
}

fn do_token_authorize<'l, 's>(
    config: &config::Config,
    realm: &schema::Realm,
    client: &schema::Client,
    user: &schema::User,
    scopes: impl Iterator<Item = &'l str> + Clone,
    state: Option<&'s str>,
) -> Result<tide::Response, OIDCError<'s>> {
    let token = token::generate_access_token(config, &realm, client, user, scopes.into_iter());

    // TODO: use api::TokenResponse here
    let response_body = serde_json::json!({
        "token": token.map_err(|e| OIDCError(OIDCErrorType::ServerError, format!("error while generating token: {:?}", e), state))?,
    });
    Ok(tide::Response::builder(200)
        .content_type(tide::http::mime::JSON)
        .body(response_body)
        .build())
}

fn do_authorize(request: Request, state: Option<&str>) -> Result<tide::Response, OIDCError> {
    let shelper = SessionHelper::new(&request);
    let realm = shelper.get_realm().map_err(|_| {
        OIDCError(
            OIDCErrorType::InvalidRequest,
            "No such realm!".to_string(),
            state,
        )
    })?;

    let make_redirect = || {
        let mut login_url = request.url().join("../v1/session/login").unwrap();
        login_url
            .query_pairs_mut()
            .clear()
            .append_pair("redirect", request.url().as_str());
        return Ok(tide::Redirect::new(login_url).into());
    };

    let qp: AuthorizeQueryParams = request
        .query()
        .map_err(|x| OIDCError(OIDCErrorType::InvalidRequest, x.to_string(), state))?;

    // collect session authentication info

    let potential_sauth = shelper
        .get_session(&request)
        .and_then(|session| shelper.get_auth_for_session(realm.id(), &session));

    let Some(user_id) = potential_sauth.and_then(|v| v.user) else {
        // if we don't have any relevant auth info, redirect to login
        return make_redirect();
    };

    let Ok(Some(user)) = realm
        .users
        .with(schema::UserID::default(), user_id)
        .first()
        .get()
    else {
        return Err(OIDCError(
            OIDCErrorType::ServerError,
            "Internal state error!".to_string(),
            state,
        ));
    };

    // verify the realm and client_id and redirect_uri

    let client = realm
        .clients
        .with(schema::Client::Shortname, &qp.client_id)
        .first()
        .get()
        .ok()
        .flatten()
        .ok_or_else(|| {
            OIDCError(
                OIDCErrorType::UnauthorizedClient,
                "Client does not exist".to_string(),
                state,
            )
        })?;

    let scopes = qp
        .scope
        .as_ref()
        .map(|slist| slist.as_str())
        .unwrap_or("")
        .split_whitespace();

    // TODO: check that redirect URI matches

    if qp.response_type == "code" {
        do_code_authorize(
            &request.state().core.config,
            &realm,
            &client,
            &user,
            scopes,
            qp.redirect_uri,
            state,
        )
    } else if qp.response_type == "token" {
        do_token_authorize(
            &request.state().core.config,
            &realm,
            &client,
            &user,
            scopes,
            state,
        )
    } else {
        Err(OIDCError(
            OIDCErrorType::UnsupportedResponseType,
            "Only code and token are understood.".to_string(),
            state,
        ))
    }
}

async fn authorize(request: Request) -> tide::Result<tide::Response> {
    #[derive(Deserialize)]
    struct State {
        state: Option<String>,
    }
    let state: Option<String> = request.query::<State>().ok().map(|x| x.state).flatten();

    match do_authorize(request, state.as_ref().map(|x| x.as_str())) {
        Ok(r) => Ok(r),
        Err(e) => Ok(e.to_response()),
    }
}

async fn do_token<'l>(mut request: Request) -> Result<tide::Response, OIDCError<'l>> {
    let shelper = SessionHelper::new(&request);
    let realm = shelper
        .get_realm()
        .map_err(|_| OIDCError(OIDCErrorType::InvalidRequest, "no such realm".into(), None))?;

    let treq: api::TokenRequestBody = request.body_form().await.map_err(|e| {
        OIDCError(
            OIDCErrorType::InvalidRequest,
            format!("could not parse form body: {e}"),
            None,
        )
    })?;

    // TODO: support HTTP basic auth for client authentication instead of treq.client_id
    let client_name = treq.client_id.ok_or(OIDCError(
        OIDCErrorType::InvalidRequest,
        "no client given".into(),
        None,
    ))?;
    let client = realm
        .clients
        .keyed((realm.id(), &client_name))
        .get()?
        .ok_or(OIDCError(
            OIDCErrorType::InvalidRequest,
            format!("unknown client name {client_name}"),
            None,
        ))?;

    if treq.grant_type == "authorization_code" {
        let Some(code) = treq.code else { todo!() };
        let code = realm
            .auth_codes
            .keyed((realm.id(), client.id(), code))
            .get()?
            .ok_or(OIDCError(
                OIDCErrorType::InvalidRequest,
                "invalid authorization code".into(),
                None,
            ))?;

        let now = std::time::SystemTime::now();
        if code.expiry < now {
            return Err(OIDCError(
                OIDCErrorType::AccessDenied,
                "expired authorization code".into(),
                None,
            ));
        }

        let user = realm
            .users
            .with(schema::UserID::default(), code.user)
            .first()
            .get()?
            .ok_or(OIDCError(
                OIDCErrorType::ServerError,
                "could not find user".into(),
                None,
            ))?;

        let access_token = token::generate_access_token(
            &request.state().core.config,
            &realm,
            &client,
            &user,
            code.scopes.as_ref().iter().map(String::as_str),
        )
        .map_err(|e| {
            OIDCError(
                OIDCErrorType::ServerError,
                format!("error signing key: {e}"),
                None,
            )
        })?;

        Ok(tide::Response::builder(200)
            .content_type(tide::http::mime::JSON)
            .body(
                serde_json::to_value(api::TokenResponse {
                    access_token: access_token.as_str(),
                    token_type: "bearer",
                    refresh_token: None,
                    scope: None,
                })
                .unwrap(),
            )
            .build())
    } else if treq.grant_type == "refresh_token" {
        let Some(rtoken) = treq.refresh_token else {
            return Err(OIDCError(OIDCErrorType::InvalidRequest, "no refresh_token given".into(), None))
        };
        let Ok((access, refresh)) =
            token::trade_refresh_token(&request.state().core.config, rtoken.as_str())
        else {
            todo!();
        };
        Ok(tide::Response::builder(200)
            .content_type(tide::http::mime::JSON)
            .body(
                serde_json::to_value(api::TokenResponse {
                    access_token: access.as_str(),
                    token_type: "bearer",
                    refresh_token: Some(refresh.as_str()),
                    scope: None,
                })
                .unwrap(),
            )
            .build())
    } else {
        Err(OIDCError(
            OIDCErrorType::InvalidRequest,
            format!("unknown grant type {}", treq.grant_type),
            None,
        ))
    }
}

async fn token(request: Request) -> tide::Result<tide::Response> {
    match do_token(request).await {
        Ok(res) => Ok(res),
        Err(e) => Ok(e.to_response()),
    }
}

async fn jwks(request: Request) -> tide::Result<tide::Response> {
    let shelper = SessionHelper::new(&request);
    let realm = shelper.get_realm()?;

    let keyinfo =
        realm
            .keys
            .get()?
            .into_iter()
            .map(|key| match key::ParsedKey::parse_from(&key)? {
                key::ParsedKey::HMAC { key_id, hmty, .. } => Ok(serde_json::json!({
                    "kid": key_id,
                    "alg": match hmty {
                        key::HMacType::Sha256 => "HS256",
                        key::HMacType::Sha512 => "HS512",
                    },
                    "hmac": true,
                })),
                key::ParsedKey::Ed25519 { key_id, keypair } => Ok(serde_json::json!({
                    "crv": "Ed25519",
                    "kid": key_id,
                    "kty": "OKP",
                    "use": "sig",

                    "x": base64::encode(keypair.public_key().as_ref()),
                })),
                key::ParsedKey::RSA { key_id, keypair } => {
                    let pubkey = keypair.public_key();
                    Ok(serde_json::json!({
                        "alg": "RS256",
                        "kid": key_id,
                        "kty": "RSA",
                        "use": "sig",

                        "e": base64::encode(pubkey.exponent().big_endian_without_leading_zero()),
                        "n": base64::encode(pubkey.modulus().big_endian_without_leading_zero()),
                    }))
                }
            });

    let jwks_response = serde_json::json!({
        "keys": keyinfo.collect::<Result<Vec<_>, UIDCError>>()?,
    });

    Ok(tide::Response::builder(200)
        .header(tide::http::headers::ACCESS_CONTROL_ALLOW_ORIGIN, "*")
        .body(jwks_response)
        .build())
}

async fn discovery_config(request: Request) -> tide::Result<tide::Response> {
    let server_config = &request.state().core.config;
    let base_url = format!(
        "{}/{}",
        server_config.base_url,
        request.param("realm").unwrap()
    );

    let config_response = serde_json::json!({
        "issuer": base_url,
        "authorization_endpoint": format!("{}/{}", base_url, AUTHORIZE_PATH),
        "token_endpoint": format!("{}/{}", base_url, TOKEN_PATH),
        "jwks_uri": format!("{}/{}", base_url, JWKS_PATH),
        "token_endpoint_auth_signing_alg_values_supported": ["EdDSA", "RS256"],
        "response_types_supported": ["code", "id_token", "token id_token"],
        "subject_types_supported": ["public"],
        "id_token_signing_alg_values_supported": ["EdDSA", "RS256"],
    });

    Ok(tide::Response::builder(200)
        .header(tide::http::headers::ACCESS_CONTROL_ALLOW_ORIGIN, "*")
        .body(config_response)
        .build())
}

pub(super) fn oidc_server(mut route: tide::Route<super::ServerStateWrapper>) {
    route.at(AUTHORIZE_PATH).get(authorize).post(authorize);
    route.at(TOKEN_PATH).post(token);
    route.at(JWKS_PATH).get(jwks);
    route.at(DISCOVERY_PATH).get(discovery_config);
}
